'use strict'

const times = require('async/times')
const crypto = require('libp2p-crypto')
const waterfall = require('async/waterfall')
const timeout = require('async/timeout')
const multihashing = require('multihashing-async')
const PeerId = require('peer-id')
const assert = require('assert')
const c = require('./constants')

const errcode = require('err-code')

class RandomWalk {
  constructor (kadDHT) {
    assert(kadDHT, 'Random Walk needs an instance of the Kademlia DHT')
    this._runningHandle = null
    this._kadDHT = kadDHT
  }

  /**
   * Start the Random Walk process. This means running a number of queries
   * every interval requesting random data. This is done to keep the dht
   * healthy over time.
   *
   * @param {number} [queries=1] - how many queries to run per period
   * @param {number} [period=300000] - how often to run the the random-walk process, in milliseconds (5min)
   * @param {number} [timeout=10000] - how long to wait for the the random-walk query to run, in milliseconds (10s)
   * @returns {void}
   */
  start (queries, period, timeout) {
    if (queries == null) { queries = 1 }
    if (period == null) { period = 5 * c.minute }
    if (timeout == null) { timeout = 10 * c.second }
    // Don't run twice
    if (this._running) { return }

    // Create running handle
    const runningHandle = {
      _onCancel: null,
      _timeoutId: null,
      runPeriodically: (fn, period) => {
        runningHandle._timeoutId = setTimeout(() => {
          runningHandle._timeoutId = null

          fn((nextPeriod) => {
            // Was walk cancelled while fn was being called?
            if (runningHandle._onCancel) {
              return runningHandle._onCancel()
            }
            // Schedule next
            runningHandle.runPeriodically(fn, nextPeriod)
          })
        }, period)
      },
      cancel: (cb) => {
        // Not currently running, can callback immediately
        if (runningHandle._timeoutId) {
          clearTimeout(runningHandle._timeoutId)
          return cb()
        }
        // Wait to finish and then call callback
        runningHandle._onCancel = cb
      }
    }

    // Start runner
    runningHandle.runPeriodically((done) => {
      this._walk(queries, timeout, () => done(period))
    }, period)
    this._runningHandle = runningHandle
  }

  /**
   * Stop the random-walk process.
   * @param {function(Error)} callback
   *
   * @returns {void}
   */
  stop (callback) {
    const runningHandle = this._runningHandle

    if (!runningHandle) {
      return callback()
    }

    this._runningHandle = null
    runningHandle.cancel(callback)
  }

  /**
   * Do the random walk work.
   *
   * @param {number} queries
   * @param {number} walkTimeout
   * @param {function(Error)} callback
   * @returns {void}
   *
   * @private
   */
  _walk (queries, walkTimeout, callback) {
    this._kadDHT._log('random-walk:start')

    times(queries, (i, cb) => {
      waterfall([
        (cb) => this._randomPeerId(cb),
        (id, cb) => timeout((cb) => {
          this._query(id, cb)
        }, walkTimeout)(cb)
      ], (err) => {
        if (err) {
          this._kadDHT._log.error('random-walk:error', err)
          return callback(err)
        }

        this._kadDHT._log('random-walk:done')
        callback(null)
      })
    })
  }

  /**
   * The query run during a random walk request.
   *
   * @param {PeerId} id
   * @param {function(Error)} callback
   * @returns {void}
   *
   * @private
   */
  _query (id, callback) {
    this._kadDHT._log('random-walk:query:%s', id.toB58String())

    this._kadDHT.findPeer(id, (err, peer) => {
      if (err.code === 'ERR_NOT_FOUND') {
        // expected case, we asked for random stuff after all
        return callback()
      }
      if (err) {
        return callback(err)
      }
      this._kadDHT._log('random-walk:query:found', err, peer)

      // wait what, there was something found? Lucky day!
      callback(errcode(new Error(`random-walk: ACTUALLY FOUND PEER: ${peer}, ${id.toB58String()}`), 'ERR_FOUND_RANDOM_PEER'))
    })
  }

  /**
   * Generate a random peer id for random-walk purposes.
   *
   * @param {function(Error, PeerId)} callback
   * @returns {void}
   *
   * @private
   */
  _randomPeerId (callback) {
    multihashing(crypto.randomBytes(16), 'sha2-256', (err, digest) => {
      if (err) {
        return callback(err)
      }
      callback(null, new PeerId(digest))
    })
  }
}

module.exports = RandomWalk
